from abcount.model.common import PKaAttribute


class PKaClassBuilder:
    @classmethod
    def build(cls, **overrides):
        """
        Return a CustomPKaAttribute class with overridden values.
        Any attributes not given in overrides will
        fall back to the parent values.
        """
        return type("CustomPKaAttribute", (PKaAttribute,), overrides)



from enum import Enum

from abcount.components.classifier import ABClassData
from abcount.model.classifier import AcidType
from abcount.model.classifier import BaseType
from abcount.model.ions import IonDefinition


class IonMatcher:
    """Matches AB data to the ion and species definitions."""

    def match_class_data(self, class_data: ABClassData) -> IonDefinition:
        # ignore difference AcidType.NONE | BaseType.NONE and None
        class_data.convert_nones_to_nulls()
        for ion_definition in IonRules:
            ion_obj = ion_definition.value
            if class_data == ion_obj.class_data:
                ion_obj.explanation = self._explanation_from_class_data(class_data)
                return ion_obj

        # If no definition is matched, it's likely a unbalanced zwitterion
        return IonDefinition(
            class_data=class_data,
            major_species_ph74_class="zwitterion",
            ion_class="zwitterion",
            explanation=self._explanation_from_class_data(class_data),
        )

    def _explanation_from_class_data(self, class_data):
        return ", ".join(self._get_list_of_defined_groups(class_data))

    def _get_list_of_defined_groups(self, class_data):
        class_data_dict = class_data.to_dict()
        return [f"{k}: {v.value}" for k, v in class_data_dict.items() if v]


class IonRules(Enum):
    I1 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.STRONG,
            acid_2_class=None,
            base_1_class=BaseType.STRONG,
            base_2_class=None,
        ),
        major_species_ph74_class="zwitterion",
        ion_class="zwitterion",
    )

    I2 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.STRONG,
            acid_2_class=None,
            base_1_class=BaseType.WEAK,
            base_2_class=None,
        ),
        major_species_ph74_class="zwitterion",
        ion_class="zwitterion",
    )

    I3 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.STRONG,
            acid_2_class=AcidType.STRONG,
            base_1_class=None,
            base_2_class=None,
        ),
        major_species_ph74_class="dianion",
        ion_class="diacid",
    )

    I4 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.STRONG,
            acid_2_class=AcidType.WEAK,
            base_1_class=None,
            base_2_class=None,
        ),
        major_species_ph74_class="dianion",
        ion_class="diacid",
    )

    I5 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.STRONG,
            acid_2_class=None,
            base_1_class=None,
            base_2_class=None,
        ),
        major_species_ph74_class="anion",
        ion_class="acid",
    )

    I6 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.WEAK,
            acid_2_class=None,
            base_1_class=None,
            base_2_class=None,
        ),
        major_species_ph74_class="anion",
        ion_class="weak acid",
    )

    I7 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.WEAK,
            acid_2_class=AcidType.WEAK,
            base_1_class=None,
            base_2_class=None,
        ),
        major_species_ph74_class="dianion",
        ion_class="weak diacid",
    )

    I8 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None,
            acid_2_class=None,
            base_1_class=BaseType.STRONG,
            base_2_class=None,
        ),
        major_species_ph74_class="cation",
        ion_class="base",
    )

    I9 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None,
            acid_2_class=None,
            base_1_class=BaseType.STRONG,
            base_2_class=BaseType.STRONG,
        ),
        major_species_ph74_class="dication",
        ion_class="dibase",
    )

    I10 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None,
            acid_2_class=None,
            base_1_class=BaseType.WEAK,
            base_2_class=None,
        ),
        major_species_ph74_class="cation",
        ion_class="weak base",
    )

    I11 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None,
            acid_2_class=None,
            base_1_class=BaseType.WEAK,
            base_2_class=BaseType.WEAK,
        ),
        major_species_ph74_class="dication",
        ion_class="weak dibase",
    )

    I12 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None,
            acid_2_class=None,
            base_1_class=BaseType.STRONG,
            base_2_class=BaseType.WEAK,
        ),
        major_species_ph74_class="dication",
        ion_class="dibase",
    )

    I13 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.WEAK,
            acid_2_class=None,
            base_1_class=BaseType.WEAK,
            base_2_class=None,
        ),
        major_species_ph74_class="zwitterion",
        ion_class="weak zwitterion",
    )

    I14 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=AcidType.WEAK,
            acid_2_class=None,
            base_1_class=BaseType.STRONG,
            base_2_class=None,
        ),
        major_species_ph74_class="zwitterion",
        ion_class="zwitterion",
    )

    I15 = IonDefinition(
        class_data=ABClassData(
            acid_1_class=None, acid_2_class=None, base_1_class=None, base_2_class=None
        ),
        major_species_ph74_class="neutral",
        ion_class="neutral",
    )



import logging

from abcount.model.common import PKaAttribute
from abcount.model.isoelectric import NetCharge
from abcount.model.isoelectric import pH

logger = logging.getLogger(__name__)


class pIPredictor:
    @staticmethod
    def predict_input(
        pka_dict: dict, pka_attribute_cls=PKaAttribute, rounding_digits=2
    ):
        if not pIPredictor.any_pkas(pka_dict, pka_attribute_cls):
            return None
        result = pIPredictor._predict_exact(
            pka_dict, pka_attribute_cls=pka_attribute_cls
        )
        return round(result, rounding_digits)

    @staticmethod
    def _predict_exact(pka_dict, pka_attribute_cls=PKaAttribute):
        # The pI is the pH where the net charge is zero.
        # Bisect only works when charges cross zero:
        # If neutrality lies between two pKa values, pI is their average.
        # If neutrality occurs at a single pKa boundary, pI equals that pKa:
        # Just take the bound on either acidic or basic
        if pIPredictor._q_interval_crosses_zero(
            *pIPredictor._predict_initial_q_bounds(pka_dict, pka_attribute_cls)
        ):
            return pIPredictor._predict_by_bisect(
                pka_dict, pka_attribute_cls=pka_attribute_cls
            )
        else:
            return pIPredictor._predict_by_bound(pka_dict, pka_attribute_cls)

    @staticmethod
    def any_pkas(pka_dict, __pka_attribute_cls__):
        pka_attributes = [
            __pka_attribute_cls__.ACID_1,
            __pka_attribute_cls__.ACID_2,
            __pka_attribute_cls__.BASE_1,
            __pka_attribute_cls__.BASE_2,
        ]
        return any([pka_dict[attr] for attr in pka_attributes])

    @staticmethod
    def _predict_by_bound(pka_dict: dict, pka_attribute_cls=PKaAttribute):
        return BoundRetriever(pka_attribute_cls).retrieve_bound(pka_dict)

    @staticmethod
    def _predict_by_bisect(pka_dict: dict, pka_attribute_cls=PKaAttribute):
        """Compute pI by finding pH where net charge = 0 using bisection.
        https://en.wikipedia.org/wiki/Bisection_method"""
        low, high = pIPredictor._initialise_low_and_high_pkas()
        q_low, q_high = pIPredictor._predict_initial_q_bounds(
            pka_dict, pka_attribute_cls
        )

        # this must be negative because of basic env
        # we assume that the interval crosses zero
        # otherwise, we cannot apply the bisect method
        if not pIPredictor._q_interval_crosses_zero(q_low, q_high):
            raise ValueError(f"Net charge does not cross zero in [{low}, {high}]")

        while (high - low) > 1e-4:  # tolerance for iteration
            mid = (low + high) / 2
            q_mid = NetChargeCalculator(pka_attribute_cls).calculate_at_pH(
                pka_dict, pH=pH(mid)
            )
            logger.debug(
                {
                    "low pH": low,
                    "mid pH": mid,
                    "high pH": high,
                    "q_low": q_low,
                    "q_mid": q_mid,
                    "q_high": q_high,
                }
            )

            if q_mid == 0:
                return mid

            # apply the bisect
            # reduce interval by probing which direction to go using q_mid.
            # q_low is positive (acidic environment protonates centres +1)
            # if q_mid is positive, reduce low interval (positive charge)
            elif q_mid > 0:
                low = mid
                q_low = q_mid
            # q_high is negative (basic environment deprotonates centres -1)
            # if q_mid is negative, reduce high interval (negative charge)
            elif q_mid < 0:
                high = mid
                q_high = q_mid

        return 0.5 * (low + high)

    @staticmethod
    def _initialise_low_and_high_pkas():
        return 0.0, 14.0

    @staticmethod
    def _predict_initial_q_bounds(pka_dict: dict, pka_attribute_cls=PKaAttribute):
        low, high = pIPredictor._initialise_low_and_high_pkas()

        q_low = NetChargeCalculator(pka_attribute_cls).calculate_at_pH(
            pka_dict, pH=pH(low)
        )  # this must be positive because of acidic env
        q_high = NetChargeCalculator(pka_attribute_cls).calculate_at_pH(
            pka_dict, pH=pH(high)
        )
        return q_low, q_high

    @staticmethod
    def _q_interval_crosses_zero(q1: float, q2: float) -> bool:
        # if multiplication is negative, then
        # the interval crosses 0
        return q1 * q2 < 0


class NetChargeCalculator:
    def __init__(self, pka_attribute_cls=PKaAttribute):
        self.__pka_attribute_cls__ = pka_attribute_cls
        self._set_attribute_lists()

    def _set_attribute_lists(self) -> None:
        self.acid_attribute_list = [
            self.__pka_attribute_cls__.ACID_1,
            self.__pka_attribute_cls__.ACID_2,
        ]
        self.base_attribute_list = [
            self.__pka_attribute_cls__.BASE_1,
            self.__pka_attribute_cls__.BASE_2,
        ]

    def calculate_at_pH(self, pka_dict: dict, pH: pH):
        nq = NetCharge()

        # Acidic charges
        for att in self.acid_attribute_list:
            pka_value = pka_dict[att]
            if pka_value:
                nq.add(ChargeCalculator.calculate_acid_charge(pka_value, pH.value))

        # Basic charges
        for att in self.base_attribute_list:
            pka_value = pka_dict[att]
            if pka_value:
                nq.add(ChargeCalculator.calculate_base_charge(pka_value, pH.value))
        return nq.value


class ChargeCalculator:
    """Calculates the fraction of protonated/deprotonated species based on
    pKa and pH."""

    ROUNDING_DIGITS = 5

    @staticmethod
    def calculate_acid_charge(pKa: float, pH: float) -> float:
        return round(-1.0 / (1.0 + 10 ** (pKa - pH)), ChargeCalculator.ROUNDING_DIGITS)

    @staticmethod
    def calculate_base_charge(pKa: float, pH: float) -> float:
        return round(1.0 / (1.0 + 10 ** (pH - pKa)), ChargeCalculator.ROUNDING_DIGITS)


class BoundRetriever:
    """Determine bound of either acidic or basic pKas in a dictionary."""

    def __init__(self, pka_attribute_cls=PKaAttribute):
        self.__pka_attribute_cls__ = pka_attribute_cls
        self._set_attribute_lists()

    def _set_attribute_lists(self) -> None:
        self.acid_attribute_list = [
            self.__pka_attribute_cls__.ACID_1,
            self.__pka_attribute_cls__.ACID_2,
        ]
        self.base_attribute_list = [
            self.__pka_attribute_cls__.BASE_1,
            self.__pka_attribute_cls__.BASE_2,
        ]

    def retrieve_bound(self, pka_dict: dict):
        acidic_pkas = [
            pka_dict[att]
            for att in self.acid_attribute_list
            if pka_dict[att] is not None
        ]
        basic_pkas = [
            pka_dict[att]
            for att in self.base_attribute_list
            if pka_dict[att] is not None
        ]
        if acidic_pkas and basic_pkas:
            raise ValueError(
                f"Both acidic and basic pKas found. Cannot determine bound. (got {pka_dict})"
            )
        elif acidic_pkas:
            return min(acidic_pkas)
        elif basic_pkas:
            return max(basic_pkas)
        raise ValueError(f"(got {pka_dict})")


if __name__ == "__main__":
    # Test
    example = {
        "pka_acid1": 3,
        "pka_acid2": 5.5,
        "pka_base1": None,
        "pka_base2": None,
    }
    print(pIPredictor.predict_input(example))



class pH:
    def __init__(self, value):
        if not value < 14 and not value > 0:
            raise ValueError(f"Invalid pH (got {value})")
        self._value = value

    @property
    def value(self):
        return self._value


class NetCharge:
    def __init__(self):
        self._value = 0.0

    def add(self, value: float):
        self._value += value

    @property
    def value(self):
        return self._value



import math

from abcount import PKaClassBuilder
from abcount.components.isoelectric import ChargeCalculator
from abcount.components.isoelectric import NetChargeCalculator
from abcount.components.isoelectric import pIPredictor
from abcount.model.isoelectric import pH


def test_pI_predict_mono():
    example = {
        "pka_acid1": 3,
        "pka_acid2": None,
        "pka_base1": 12,
        "pka_base2": None,
    }
    pI = pIPredictor.predict_input(example)
    assert math.isclose(pI, 7.5, abs_tol=1e-1)


def test_pI_predict_diacid():
    example = {
        "pka_acid1": 3,
        "pka_acid2": 5.5,
        "pka_base1": None,
        "pka_base2": None,
    }
    pI = pIPredictor.predict_input(example)
    assert math.isclose(pI, 3, abs_tol=1e-1)


def test_pI_predict_dibasic():
    example = {
        "pka_acid1": None,
        "pka_acid2": None,
        "pka_base1": 12,
        "pka_base2": 8.5,
    }
    pI = pIPredictor.predict_input(example)
    assert math.isclose(pI, 12, abs_tol=1e-1)


def test_pI_predict_di_and_mono():
    example = {
        "pka_acid1": 3,
        "pka_acid2": None,
        "pka_base1": 12,
        "pka_base2": 8.5,
    }
    pI = pIPredictor.predict_input(example)
    assert math.isclose(pI, 10.25, abs_tol=1e-1)


def test_pI_predict_di_di_custom():
    CustomPKaAttribute = PKaClassBuilder.build(
        ACID_1="my_pka_acid1",
        BASE_1="my_pka_base1",
        ACID_2="my_pka_acid2",
        BASE_2="my_pka_base2",
    )
    example = {
        "my_pka_acid1": 3,
        "my_pka_base1": 5.5,
        "my_pka_acid2": 12,
        "my_pka_base2": 8.5,
    }
    pI = pIPredictor.predict_input(example, CustomPKaAttribute)
    assert math.isclose(pI, 7, abs_tol=1e-1)


def test_pI_predict_di_di():
    example = {
        "pka_acid1": 3,
        "pka_acid2": 5.5,
        "pka_base1": 12,
        "pka_base2": 8.5,
    }
    pI = pIPredictor.predict_input(example)
    assert math.isclose(pI, 7, abs_tol=1e-1)


def test_pI_predict_weak_leaking():
    example = {
        "pka_acid1": 10,
        "pka_acid2": None,
        "pka_base1": 2.4,
        "pka_base2": 2.3,
    }
    pI = pIPredictor.predict_input(example)
    assert math.isclose(pI, 6.33, abs_tol=1e-1)


def test_netcharge_calc():
    example = {
        "pka_acid1": 3,
        "pka_acid2": None,
        "pka_base1": 11,
        "pka_base2": None,
    }
    nq = NetChargeCalculator().calculate_at_pH(example, pH=pH(7.4))
    assert math.isclose(nq, 0.0, abs_tol=1e-2)


def test_no_groups():
    example = {
        "pka_acid1": None,
        "pka_acid2": None,
        "pka_base1": None,
        "pka_base2": None,
    }
    pI = pIPredictor.predict_input(example)
    assert pI is None


def test_netcharge_custom_calc():
    example = {"pka_1": 3, "pka_2": None, "pkb_1": 11, "pkb_2": None}
    CustomPKaAttribute = PKaClassBuilder.build(
        ACID_1="pka_1", ACID_2="pka_2", BASE_1="pkb_1", BASE_2="pkb_2"
    )
    nq = NetChargeCalculator(CustomPKaAttribute).calculate_at_pH(example, pH=pH(7.4))
    assert math.isclose(nq, 0.0, abs_tol=1e-2)


def test_acid_charge_half_species():
    q = ChargeCalculator.calculate_acid_charge(pKa=7.4, pH=7.4)
    assert math.isclose(q, -0.5, abs_tol=1e-2)


def test_acid_charge_proto_species():
    q = ChargeCalculator.calculate_acid_charge(pKa=3, pH=7.4)
    assert math.isclose(q, -1.0, abs_tol=1e-2)


def test_acid_charge_deproto_species():
    q = ChargeCalculator.calculate_acid_charge(pKa=11, pH=7.4)
    assert math.isclose(q, 0.0, abs_tol=1e-2)


def test_base_charge_half_species():
    q = ChargeCalculator.calculate_base_charge(pKa=7.4, pH=7.4)
    assert math.isclose(q, 0.5, abs_tol=1e-2)


def test_base_charge_proto_species():
    q = ChargeCalculator.calculate_base_charge(pKa=11, pH=7.4)
    assert math.isclose(q, 1.0, abs_tol=1e-2)


def test_base_charge_deproto_species():
    q = ChargeCalculator.calculate_base_charge(pKa=3, pH=7.4)
    assert math.isclose(q, 0.0, abs_tol=1e-2)
